/**
 * firmwareXTL
 * By: Xiao-Wen Terry Liu (6730047)
 * This is a firmware for the Lego Mindstorm robot partially based on
 * the code written by Professor Jacky Baltes at University of Manitoba.
 */


/**
 * ======================================================================
 * GLOBAL CONSTANTS
 * ======================================================================
 */

#define DEBUG
#define MY_PORT 2
#define DEST_HOST 0x8
#define DEST_PORT 0x7
#define DEST_PORT_2 0x8
#define DEST_ADDR (DEST_HOST << 4 | DEST_PORT)
#define DEST_ADDR_2 (DEST_HOST << 4 | DEST_PORT_2)
#define LEN 32

//Calibration constants
#define NORMAL_SPEED (MAX_SPEED)
#define TURN_SPEED   ((2*MAX_SPEED)/2)
#define FWD_V 70
#define REV_V 90
#define FWD_T 30  //was 12
#define REV_T 30  //was 12


#define DIR_FWD fwd							//Map FWD direction to our robot's hardware
#define DIR_REV rev							//Map REV direction to our robot's hardware

//Msg constants
#define IR_RANGE		0					//0=Short range, 1=Long range
#define ROBOTS_NUM		8					//Number of robots
#define MSG_BOM 		0x80				//Beginning Of Message = 128 (-8, 0) or (1000 0000)

#define LEFT 			0					//The sign to turn Left
#define RIGHT 			1					//The sign to turn right


//Commands for robot
#define CMD_FWD_KICK 			8		//A forward kick (0, -8)
#define CMD_SPIN_KICK_LEFT		0x98		//Spin kick to the left (-7,-8) 1001 1000
#define CMD_SPIN_KICK_RIGHT		0x78		//Spin kick to the right (+7,-8) 0111 1000
#define CMD_IGNORE 			0xFF81		//The ignore command (-8, 1)


/**
 * ======================================================================
 * Libraries
 * ======================================================================
 */

#include <conio.h>							//console IO
#include <unistd.h>
#include <lnp/sys/lnp.h>
#include <lnp/sys/lnp-logical.h>
#include <lnp/lnp-logical.h>
#include <stdlib.h>
#include <string.h>

#include <dsensor.h>
#include <dkey.h>							//for dkey
#include <dmotor.h>
#include <dsound.h>
#include <sys/tm.h>


/**
 * ======================================================================
 * GLOBAL VARIABLES
 * ======================================================================
 */
static volatile signed char gdata[10];			//global: the message
static volatile unsigned char id;				//the ID of the robot
static unsigned char rv;						//flag that we received a msg


/**
 * ======================================================================
 * Prototypes
 * ======================================================================
 */
wakeup_t prgmKeyCheck(wakeup_t data);
wakeup_t cmdReceived(wakeup_t data);
wakeup_t cmdOrKey(wakeup_t dummy);
void rx_msg( const unsigned char * data, unsigned char len );
int passCRC();
void motor_drive();


/**
 * ======================================================================
 * The MAIN function
 * ======================================================================
 */
int main(int argc,char *argv[]) {
	//Local variables
	tid_t process_id = -1;						//the process ID


	//Initializing code
	lnp_logical_init();							//May not have to call

	lnp_logical_range( IR_RANGE );				//Set the IR range

	lnp_integrity_set_handler(rx_msg);			//Set the handle to incoming msg function

	memset((void *)gdata, 0, sizeof(gdata) );	//Set the memory

	id = 0;										//Set the id default 0



     //Normal execution code
	while(1) {
		cputs("wait");							//Display "wait"
		wait_event(cmdOrKey, 0);				//Suspend program until triggered

		//If the program key was pressed
		if(prgmKeyCheck(0)) {
			//Ensure proper termination of process
			if (process_id != -1) {
				kill(process_id);
			}//end if

			//Try to find a robot ID
			id = (++id)%ROBOTS_NUM;				//Try next ID

			cputw(id);							//Display ID
			msleep(500);						//Keep display for 1/2 second
		} else {
			//It wasn't the program key, it was the received command

			//Check if it wasn't the ignore command
			if (gdata[id] != CMD_IGNORE) {
				//Ensure proper termination of process
				if (process_id != -1) {
					kill(process_id);
				}//end if

				//Create a process
				process_id = execi((void *)&motor_drive,0,0,PRIO_NORMAL,DEFAULT_STACK_SIZE);
			}//end if
		}//end if
	}//end while



	return 0;		//Normal exit of program
}//end main


/**
 * ======================================================================
 * The SUPPORTING function
 * ======================================================================
 */

/**
 * Handles incoming messages
 */
void rx_msg( const unsigned char * data, unsigned char len ) {
	int i;					//loop counter
	for( i=0; i< len-1; i++) {
		gdata[i] = data[i+1]; //copy the passed data to global data
	}//end for
	//gdata[len] = data[0];	//Move the header to the last byte (easier referencing later)

	rv = 1;					//set the rcv flag
}//end rx_msg


/**
 * Checks if the program key is pressed
 */
wakeup_t prgmKeyCheck(wakeup_t data) {
	wakeup_t value = 0;		//return value

	//Only if the program key was pressed.
	if ( dkey == KEY_PRGM) {
		value = 1;
	}
	return value;
}//end prgmKeyCheck


/**
 * Checks if a receive command was received
 */
wakeup_t cmdReceived(wakeup_t data) {
	//Assign the return value to be whatever was the rv flag
	wakeup_t value = rv;

	rv = 0;					//reset the rcv flag
	return value;
}//end cmdReceived


/**
 * Whether the command received or key was pressed
 */
wakeup_t cmdOrKey(wakeup_t dummy) {
	return (cmdReceived(0) || prgmKeyCheck(0));
}//end cmdOrKey


/**
 * Does a simple check on the message integrity
 */
int passChecksum() {
	int value = 0;			//the return value
	int count = 0;			//the counter
	int i;					//temp. loop counter

	//Loop for every command in the message
	for ( i = 0; i < 8;i++) {
		count = count + gdata[i];

	}//end for

	//If pass the check, then change return value
	if ( (count & 0x0007) == (gdata[8] & 0x0007) ){
		value = 1;			//Change return value to pass
	}//end if

	return value;
}//end passChecksum


/**
 * The function that handles the motor functions
 */
void motor_drive() {
	int t;					//the sleep time
	int temp;
	char cmd;				//the command

	cmd = gdata[id];		//extract our command.




	if ( passChecksum() ) {
		if (cmd != CMD_IGNORE) {
			/*
			cputw(cmd);
			sleep(1);
			motor_a_speed(0);
			motor_c_speed(0);
			*/

			if ( cmd == CMD_FWD_KICK ) {
				//Do kick
				motor_a_dir(DIR_FWD);
				motor_c_dir(DIR_FWD);
				motor_a_speed(NORMAL_SPEED);
				motor_c_speed(NORMAL_SPEED);
				//t = 0x30 * FWD_V;
				t = 500;
				msleep(t);

				//Then stop
				motor_a_speed(0);
				motor_c_speed(0);
				motor_a_dir(brake);			//brake 1st motor
				motor_c_dir(brake);			//brake 1st motor

			} else if ( cmd == CMD_SPIN_KICK_LEFT) {
				//Do kick
				motor_a_dir(DIR_FWD);
				motor_c_dir(DIR_REV);
				motor_a_speed(NORMAL_SPEED);
				motor_c_speed(NORMAL_SPEED);

				t = 1000;

				msleep(t);

				//Then stop
				motor_a_speed(0);
				motor_c_speed(0);
			} else if ( cmd == CMD_SPIN_KICK_RIGHT) {
				//Do kick
				motor_a_dir(DIR_REV);
				motor_c_dir(DIR_FWD);
				motor_a_speed(NORMAL_SPEED);
				motor_c_speed(NORMAL_SPEED);

				t = 1000;

				msleep(t);

				//Then stop
				motor_a_speed(0);
				motor_c_speed(0);
			}
			else {
				//Assume it's a regular motor command

				//Do turning using 1st 4 bytes

				/*
				lcd_int(2);
				msleep(500);
				lcd_int(cmd);
				msleep(500);
				*/

				//Check direction
				if ( (cmd >> 7) & 0x01 ) {
					//Turn to the left
					motor_a_dir(DIR_FWD);
					motor_c_dir(DIR_REV);

					//t = ( ( (cmd >> 4) & 0x0F) & 0x07)  * REV_T; ////for not using 2's complement
					temp =( ( (~( (cmd >> 4) | 0xF0 ))) +0x01) & 0xFF;
					t =  temp * REV_T;

					//cputw(temp);
					//msleep(1000);

				} else {
					//Turn to the right
					motor_a_dir(DIR_REV);
					motor_c_dir(DIR_FWD);

					t = ( (cmd >> 4) & 0x0F) * FWD_T;

				}//end if

				//Turn!!
				motor_a_speed(TURN_SPEED);
				motor_c_speed(TURN_SPEED);

				msleep(t);

				//Then stop
				motor_a_speed(0);
				motor_c_speed(0);

				//Do movement using last 4 bytes

				//Check direction to travel by the fourth bit
				if ( (cmd >> 3) & 0x01 ) {
					//It's a 1, thus negative, or reverse
					motor_a_dir(DIR_REV);
					motor_c_dir(DIR_REV);

					//t = ( (cmd & 0x0F) & 0x07) * REV_V; //for not using 2's complement
					temp =( (~(cmd | 0xF0)+0x01) ) & 0xFF;
					t = temp  * REV_V; //for 2's complement


				} else {
					//It's a 0, thus positive, or forward
					motor_a_dir(DIR_FWD);
					motor_c_dir(DIR_FWD);

					t = (cmd & 0x0F ) * FWD_V;
				}//end if

				//Move!!
				motor_a_speed(NORMAL_SPEED);
				motor_c_speed(NORMAL_SPEED);

				msleep(t);

				//Then stop
				motor_a_speed(0);
				motor_c_speed(0);

			}//end if

		}//end if
	}//end if


}//end motor_drive

